import datetime
import logging
import numbers
import os
from itertools import zip_longest
from typing import Any
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Sequence
from typing import Union

import h5py

logger = logging.getLogger(__name__)


def mca_data_to_spec_string(
    mca: Sequence[float],
    title: Optional[str] = None,
    filename: Optional[str] = None,
    date: Optional[Union[str, datetime.datetime, datetime.date]] = None,
    calibration: Optional[Sequence[float]] = None,
    detector_name: Optional[str] = None,
    metadata: Optional[Dict] = None,
) -> str:
    """Official file format specs: https://certif.com/downloads/css_docs/spec_man.pdf

    :param mca: MCA spectrum
    :param title: scan title
    :param filename: name by which the file will be created
    :param date: start date of the scan
    :param calibration: MCA calibration as the sequence [zero, gain, quad],
                        all of them optional so that
                        energy = zero + gain*channels + quad*channels^2
    :param detector_name: name of the MCA detector
    :param metadata: saved as comments
    :returns: SPEC serialized string
    """
    nchan_per_line = 16
    nchan_tot = len(mca)

    if title is None:
        title = "ct"
    if filename is None:
        filename = "unspecified"
    if detector_name is None:
        detector_name = "MCA0"
    date = spectimeformat(date)
    calib = [0, 1, 0]  # zero, gain, quad
    if calibration:
        if len(calibration) > 3:
            raise ValueError("MCA calibration requires 3 only coefficients")
        calib = [
            p if p is not None else pdefault
            for p, pdefault in zip_longest(calibration, calib)
        ]

    header = [f"#F {filename}", f"#D {date}", "", f"#S {title}", f"#D {date}"]
    if metadata:
        for k, v in metadata.items():
            header.append(f"#C {k} = {v}")
    header.append("#N 1")
    header.append(f"#@MCA {nchan_per_line}C")
    header.append(f"#@CHANN {nchan_tot} 0 {nchan_tot-1} 1")
    header.append(f"#@CALIB {' '.join(map(str, calib))}")
    header.append("#@MCA_NB 1")
    header.append(f"#L {detector_name}")

    mcastring = "\n".join(header)

    mcastring += "\n@A"
    if isinstance(mca[0], numbers.Integral):
        fmt = " %d"
    else:
        # fmt = " %.4f"
        fmt = " %.8g"
    for idx in range(0, nchan_tot, nchan_per_line):
        if idx + nchan_per_line - 1 < nchan_tot:
            for i in range(0, nchan_per_line):
                mcastring += fmt % mca[idx + i]
            if idx + nchan_per_line != nchan_tot:
                mcastring += "\\"
        else:
            for i in range(idx, nchan_tot):
                mcastring += fmt % mca[i]
        mcastring += "\n"
    return mcastring


def save_as_spec(filename: str, mca: Sequence[float], **kwargs) -> None:
    dirname = os.path.dirname(filename)
    if dirname:
        os.makedirs(dirname, exist_ok=True)
    content = mca_data_to_spec_string(mca, filename=filename, **kwargs)
    with open(filename, "w") as f:
        f.write(content)


class SpecScanData(NamedTuple):
    scan_name: str
    scan_title: str
    start_time: str
    scan_data: Dict[str, list]
    n_points: int

    @property
    def scan_number(self) -> int:
        return int(float(self.scan_name))

    @property
    def spec_title(self) -> str:
        return f"#S {self.scan_number} {self.scan_title}"


def extract_scan_data(
    h5file: h5py.File,
    scan_name: str,
    counter_names: List[str],
    acq_time: Optional[float],
) -> SpecScanData:
    """
    Extract counters and start_time for a single scan.
    Trim all counters to the same minimum length.
    Handle missing counters and timer/acq_time injection.
    """
    scan = h5file[scan_name]
    _ = scan["end_time"]
    measurement = scan["measurement"]
    force_counters = bool(counter_names)
    if not force_counters:
        counter_names = list(measurement)

    scan_data: Dict[str, list] = {}

    # Extract and format start_time
    dset = scan.get("start_time")
    if dset is not None:
        start_time = spectimeformat(dset[()])
    else:
        start_time = spectimeformat(None)
        logger.warning(f"Missing start_time for scan {scan_name}, using dummy time.")

    # Extract scan title
    dset = scan.get("title")
    if dset is not None:
        scan_title = asstr(dset[()])
    else:
        scan_title = "NISscan"

    # Extract counters
    for name in counter_names:
        if name in measurement:
            dset = measurement[name]
            if dset.ndim != 1:
                if force_counters:
                    logger.warning(
                        f"Counter '{name}' is not 1D in scan {scan_name}, skipping."
                    )
            else:
                scan_data[name] = dset[()]
        else:
            logger.warning(f"Counter '{name}' missing in scan {scan_name}, skipping.")

    # Determine minimum length and trim
    if scan_data:
        min_len = min(len(v) for v in scan_data.values())
        for k in scan_data.keys():
            scan_data[k] = scan_data[k][:min_len]
    else:
        logger.warning(f"No counters found for scan {scan_name}. Skipping this scan.")
        return SpecScanData(
            scan_name=scan_name,
            scan_title=scan_title,
            start_time=start_time,
            scan_data={},
            n_points=0,
        )

    # Handle acquisition time
    if "timer" in scan_data:
        scan_data["Seconds"] = scan_data.pop("timer")
    elif acq_time is not None:
        scan_data["Seconds"] = [acq_time] * min_len
    else:
        dset = h5file.get(f"/{scan_name}/scan_parameters/count_time")
        if dset is not None:
            acq_time = dset[()]
        else:
            acq_time = 0
            logger.warning(
                f"No 'acq_time' parameter or 'timer' counter found for scan {scan_name}, inserting zeros."
            )
        scan_data["Seconds"] = [acq_time] * min_len

    return SpecScanData(
        scan_name=scan_name,
        scan_title=scan_title,
        start_time=start_time,
        scan_data=scan_data,
        n_points=min_len,
    )


def scan_data_to_spec_string(scan: SpecScanData) -> str:
    lines = []
    lines.append(scan.spec_title)
    lines.append(f"#D {scan.start_time}")
    lines.append(f"#T {scan.scan_data['Seconds'][0]}  (Seconds)")

    keys = list(scan.scan_data)
    lines.append("#L " + "  ".join(keys))

    for k in range(scan.n_points):
        row = [str(scan.scan_data[name][k]) for name in keys]
        lines.append(" ".join(row) + "")

    lines.append("#C")
    lines.append("")
    lines.append("")
    return "\n".join(lines)


def save_as_spec2(filename: str, scans: List[SpecScanData]) -> str:
    """
    Write the extracted scan data into an ASCII file in SPEC format.
    """
    filename = os.path.abspath(filename)
    dirname = os.path.dirname(filename)
    if dirname:
        os.makedirs(dirname, exist_ok=True)

    existing_content = ""
    if os.path.exists(filename):
        with open(filename, "r") as f:
            existing_content = f.read()
    else:
        with open(filename, "a") as f:
            date = spectimeformat(None)
            f.write(f"#F {filename}\n")
            f.write(f"#D {date}\n")
            f.write("\n")

    with open(filename, "a") as f:
        for scan in scans:
            if scan.n_points == 0:
                continue
            if scan.spec_title in existing_content:
                # Scan was already saved
                continue
            content = scan_data_to_spec_string(scan)
            f.write(content)
    return filename


def spectimeformat(date: Union[datetime.datetime, str, bytes, None]) -> str:
    if date is None:
        date = datetime.datetime.now()
    if not isinstance(date, datetime.datetime):
        date = datetime.datetime.fromisoformat(asstr(date))
    return date.strftime("%a %b %d %H:%M:%S %Y")


def asstr(data: Union[str, bytes]) -> Any:
    if isinstance(data, bytes):
        return data.decode()
    return data



import numpy

from ..tasks import io


def test_io_spec():
    time_iso = "2025-08-27T13:53:01.817551"
    time_spec = "Wed Aug 27 13:53:01 2025"

    mca = numpy.arange(6)
    mca_string = io.mca_data_to_spec_string(mca, date=time_iso)
    expected = [
        "#F unspecified",
        f"#D {time_spec}",
        "",
        "#S ct",
        f"#D {time_spec}",
        "#N 1",
        "#@MCA 16C",
        "#@CHANN 6 0 5 1",
        "#@CALIB 0 1 0",
        "#@MCA_NB 1",
        "#L MCA0",
        "@A 0 1 2 3 4 5",
        "",
    ]
    assert mca_string.split("\n") == expected

    mca = numpy.arange(16)
    mca_string = io.mca_data_to_spec_string(mca, date=time_iso)
    expected = [
        "#F unspecified",
        f"#D {time_spec}",
        "",
        "#S ct",
        f"#D {time_spec}",
        "#N 1",
        "#@MCA 16C",
        "#@CHANN 16 0 15 1",
        "#@CALIB 0 1 0",
        "#@MCA_NB 1",
        "#L MCA0",
        "@A 0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15",
        "",
    ]
    assert mca_string.split("\n") == expected

    mca = numpy.arange(32)
    mca_string = io.mca_data_to_spec_string(mca, date=time_iso)
    expected = [
        "#F unspecified",
        f"#D {time_spec}",
        "",
        "#S ct",
        f"#D {time_spec}",
        "#N 1",
        "#@MCA 16C",
        "#@CHANN 32 0 31 1",
        "#@CALIB 0 1 0",
        "#@MCA_NB 1",
        "#L MCA0",
        "@A 0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15\\",
        " 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31",
        "",
    ]
    assert mca_string.split("\n") == expected

    mca = numpy.arange(33)
    mca_string = io.mca_data_to_spec_string(mca, date=time_iso)
    expected = [
        "#F unspecified",
        f"#D {time_spec}",
        "",
        "#S ct",
        f"#D {time_spec}",
        "#N 1",
        "#@MCA 16C",
        "#@CHANN 33 0 32 1",
        "#@CALIB 0 1 0",
        "#@MCA_NB 1",
        "#L MCA0",
        "@A 0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15\\",
        " 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31\\",
        " 32",
        "",
    ]
    assert mca_string.split("\n") == expected

    mca = numpy.arange(33)
    mca_string = io.mca_data_to_spec_string(
        mca, date=time_iso, metadata={"a": 1, "b": 2}
    )
    expected = [
        "#F unspecified",
        f"#D {time_spec}",
        "",
        "#S ct",
        f"#D {time_spec}",
        "#C a = 1",
        "#C b = 2",
        "#N 1",
        "#@MCA 16C",
        "#@CHANN 33 0 32 1",
        "#@CALIB 0 1 0",
        "#@MCA_NB 1",
        "#L MCA0",
        "@A 0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15\\",
        " 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31\\",
        " 32",
        "",
    ]
    assert mca_string.split("\n") == expected


def test_io_spec2():
    scan = io.SpecScanData(
        "1.1",
        "NISscan",
        "Wed Aug 27 13:53:01 2025",
        scan_data={"diode1": [1, 2], "diode2": [3, 4], "Seconds": [0.5, 0.5]},
        n_points=2,
    )
    spec_string = io.scan_data_to_spec_string(scan)
    expected = """#S 1 NISscan
#D Wed Aug 27 13:53:01 2025
#T 0.5  (Seconds)
#L diode1  diode2  Seconds
1 3 0.5
2 4 0.5
#C

"""
    assert spec_string == expected



from typing import List

import h5py
import numpy

from ...tasks.hdf5_to_spec2 import Hdf5ToSpec2


def test_hdf5_to_spec2(tmp_path):
    raw_filename = tmp_path / "RAW_DATA" / "bliss_dataset.h5"
    raw_filename.parent.mkdir()
    raw_output_filename = tmp_path / "PROCESSED_DATA" / "bliss_dataset.mca"
    filename = str(raw_filename)
    output_filename = str(raw_output_filename)

    time_iso = "2025-08-27T13:53:01.817551"
    time_spec = "Wed Aug 27 13:53:01 2025"

    nscans = 3
    nchannels = 10
    with h5py.File(filename, "w") as nxroot:
        for scan in range(1, nscans + 1):
            nxroot[f"/{scan}.1/start_time"] = time_iso
            nxroot[f"/{scan}.1/title"] = "NISscan"
            nxroot[f"/{scan}.1/measurement/diode1"] = numpy.full((nchannels,), scan)
            nxroot[f"/{scan}.1/end_time"] = time_iso

    def check_scan_content(scans: List[int], scan_acq_times: List[float]) -> List[str]:
        with open(task.outputs["output_filename"], "r") as f:
            lines = [s.rstrip() for s in f.readlines()]

        expected = [f"#F {output_filename}", lines[1], ""]
        if not scan_acq_times:
            scan_acq_times = [0] * len(scans)
        for scan, acq_time in zip(scans, scan_acq_times):
            expected.append(f"#S {scan} NISscan")
            expected.append(f"#D {time_spec}")
            expected.append(f"#T {acq_time}  (Seconds)")
            expected.append("#L diode1  Seconds")
            for _ in range(nchannels):
                expected.append(f"{scan} {acq_time}")
            expected.append("#C")
            expected.append("")

        assert lines == expected

    all_scans_numbers = list(range(1, nscans + 1))

    # Test fresh file
    inputs = {
        "filename": filename,
        "output_filename": output_filename,
        "scan_numbers": all_scans_numbers[:-1],
    }
    task = Hdf5ToSpec2(inputs=inputs)
    task.run()
    check_scan_content(all_scans_numbers[:-1], [])

    # Test no-duplicates and append
    inputs = {"filename": filename, "output_filename": output_filename}
    task = Hdf5ToSpec2(inputs=inputs)
    task.run()
    check_scan_content(all_scans_numbers, [])

    # Test scan_acq_times
    raw_output_filename.unlink()
    inputs = {
        "filename": filename,
        "output_filename": output_filename,
        "scan_acq_times": all_scans_numbers,
    }
    task = Hdf5ToSpec2(inputs=inputs)
    task.run()
    check_scan_content(all_scans_numbers, all_scans_numbers)

    # Test scan_acq_times ignored
    scan_acq_times = []
    with h5py.File(filename, "a") as nxroot:
        for scan in range(1, nscans + 1):
            acq_time = scan / 10
            scan_acq_times.append(acq_time)
            nxroot[f"/{scan}.1/measurement/timer"] = numpy.full((nchannels,), acq_time)

    raw_output_filename.unlink()
    inputs = {
        "filename": filename,
        "output_filename": output_filename,
        "scan_acq_times": all_scans_numbers,
    }
    task = Hdf5ToSpec2(inputs=inputs)
    task.run()
    check_scan_content(all_scans_numbers, scan_acq_times)



from typing import List
from typing import Optional

from blissdata.h5api import dynamic_hdf5
from ewokscore.model import BaseInputModel
from ewokscore.task import Task
from pydantic import Field

from .io import SpecScanData
from .io import extract_scan_data
from .io import save_as_spec2


class Inputs(BaseInputModel):
    filename: str = Field(
        ...,
        description="Path to the HDF5 file. Can be absolute or relative.",
        examples=[
            "/data/visitor/hc6232/id14/20250716/PROCESSED_DATA/MK103_235K_IV/session_HRM2_22Jul25.h5",
            "session_HRM2_01Apr25.h5",
        ],
    )

    output_filename: str = Field(
        ...,
        description="Output filename for the extracted ASCII data. Can be absolute or relative.",
        examples=["jul16.asc", "/home/user/myname/output.asc"],
    )

    scan_numbers: Optional[List[int]] = Field(
        None,
        description="List of scans to extract.",
        examples=[[29, 30, 39, 40, 49, 50]],
    )

    counter_names: Optional[List[str]] = Field(
        None,
        description="List of counters/motors to extract.",
        examples=[["bin2", "bout2", "eib1", "pico2", "df", "di", "epoch", "timer"]],
    )

    scan_acq_times: Optional[List[float]] = Field(
        None,
        description=(
            "Acquisition time per point. Either a single-element list (applied to all scans), "
            "or a list matching the length of scan_numbers. When not provided try using the 'timer' counter."
        ),
        examples=[[8], [8, 8, 10]],
    )

    retry_timeout: Optional[float] = Field(
        None, description="Timeout for trying to read from the HDF5 file in seconds."
    )

    retry_period: Optional[float] = Field(
        None,
        description="Retry period for trying to read from the HDF5 file in seconds.",
    )


class Hdf5ToSpec2(Task, input_model=Inputs, output_names=["output_filename"]):
    """Convert Bliss HDF5 scans to SPEC files, multiple scans per file and only 1D counters."""

    def run(self):
        filename = self.inputs.filename
        scan_numbers = self.get_input_value("scan_numbers", [])
        counter_names = self.get_input_value("counter_names", [])
        scan_acq_times = self.get_input_value("scan_acq_times", [])
        retry_timeout = self.get_input_value("retry_timeout", 0)
        retry_period = self.get_input_value("retry_period", 1)

        scans: List[SpecScanData] = []
        with dynamic_hdf5.File(
            filename, retry_timeout=retry_timeout, retry_period=retry_period
        ) as nxroot:
            if scan_numbers:
                scan_names = [f"{scannr}.1" for scannr in scan_numbers]
            else:
                scan_names = list(nxroot)

            if scan_acq_times:
                if len(scan_acq_times) == 1:
                    scan_acq_times = scan_acq_times * len(scan_names)
                elif len(scan_acq_times) != len(scan_names):
                    raise ValueError(
                        "Length of 'scan_acq_times' and 'scan_numbers' must be the same"
                    )
            else:
                scan_acq_times = [None] * len(scan_names)

            for scan_name, acq_time in zip(scan_names, scan_acq_times):
                scan = extract_scan_data(nxroot, scan_name, counter_names, acq_time)
                scans.append(scan)

        self.outputs.output_filename = save_as_spec2(self.inputs.output_filename, scans)
